"""Client exceptions."""

import asyncio


class ClientError(Exception):
    """Base exception for clients."""


class ClientConnectionError(ClientError):
    """Exception for client connection errors."""


class ClientTimeoutError(asyncio.TimeoutError, ClientConnectionError):
    """Exception for client connection errors caused by timeouts."""


class ClientResponseError(ClientError):
    """Exception for client response errors."""


class CommandError(ClientError):
    """Base exception for errors caused by sending commands or requests."""


class LoginFailedError(CommandError):
    """Login failed."""


class LoginRequiredError(CommandError):
    """Login is required to perform this command."""



"""IConfiguration.OpenFilter method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional


@dataclass
class OpenFilter:
    """IConfiguration.OpenFilter method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional["OpenFilter.Params"] = field(default=None)
    return_value: Optional[int] = field(
        default=None,
        metadata={
            "name": "return",
        },
    )

    @dataclass
    class Params:
        """IConfiguration.OpenFilter method parameters."""

        object_types: Optional[List[str]] = field(
            default=None,
            metadata={
                "name": "ObjectType",
                "wrapper": "Objects",
            },
        )

        xpath: Optional[str] = field(
            default=None,
            metadata={
                "name": "XPath",
            },
        )



"""RGB load object."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Tuple

from .child_object import ChildObject
from .location_object import LocationObject


@dataclass
class RGBLoadBase(ChildObject, LocationObject):
    """RGB load base class."""

    class ColorType(Enum):
        RGB = "RGB"
        RGBW = "RGBW"
        HSL = "HSL"
        HSIC = "HSIC"
        CCT = "CCT"
        COLOR_CHANNEL = "Color Channel"

    color_type: ColorType = field(
        metadata={
            "name": "ColorType",
        }
    )

    min_temp: int = field(
        metadata={
            "name": "MinTemp",
        }
    )

    max_temp: int = field(
        metadata={
            "name": "MaxTemp",
        }
    )

    hsl: Optional[Tuple[int, int, int]] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    rgb: Optional[Tuple[int, int, int]] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    rgbw: Optional[Tuple[int, int, int, int]] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    level: Optional[int] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    color_temp: Optional[int] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)

    @property
    def is_rgb(self) -> bool:
        """Return True if the load is an RGB(W) load."""
        return self.color_type in (
            self.ColorType.RGB,
            self.ColorType.RGBW,
            self.ColorType.HSL,
        )

    @property
    def is_cct(self) -> bool:
        """Return True if the load is a CCT load."""
        return self.color_type == self.ColorType.CCT



"""Base class for child device objects."""

from dataclasses import dataclass

from .child_object import ChildObject
from .custom_device import CustomDevice


@dataclass
class ChildDevice(ChildObject, CustomDevice):
    """Base class for child device objects."""



"""Base class for all objects."""

from dataclasses import dataclass, field

from xsdata.models.datatype import XmlDateTime


@dataclass
class SystemObject:
    """Base class for all objects."""

    id: int = field(
        metadata={
            "name": "VID",
            "type": "Attribute",
        }
    )

    master_id: int = field(
        metadata={
            "name": "Master",
            "type": "Attribute",
        }
    )

    mtime: XmlDateTime = field(
        metadata={
            "name": "MTime",
            "type": "Attribute",
        }
    )

    name: str = field(
        metadata={
            "name": "Name",
        }
    )

    note: str = field(
        metadata={
            "name": "Note",
        }
    )

    model: str = field(
        metadata={
            "name": "Model",
        }
    )

    display_name: str = field(
        metadata={
            "name": "DName",
        }
    )



"""Somfy RS-485 SDN 2.0 blind."""

from dataclasses import dataclass

from .blind_base import BlindBase
from .child_device import ChildDevice


@dataclass
class SomfyRS485ShadeChild(BlindBase, ChildDevice):
    """Somfy RS-485 SDN 2.0 blind."""

    class Meta:
        name = "Somfy.RS-485_Shade_CHILD"



"""Base class for system objects in an area."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass
class LocationObject(SystemObject):
    """Base class for system objects in an area."""

    area_id: int = field(
        metadata={
            "name": "Area",
        }
    )

    location: str = field(
        metadata={
            "name": "Location",
        }
    )



"""Relay blind object."""

from dataclasses import dataclass

from .blind_base import BlindBase
from .location_object import LocationObject


@dataclass
class RelayBlind(BlindBase, LocationObject):
    """Relay blind object."""



"""OmniSensor object."""

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Union

from .child_object import ChildObject
from .sensor import Sensor


@dataclass
class Formula:
    """OmniSensor type conversion information."""

    return_type: str = field(
        metadata={
            "name": "ReturnType",
            "type": "Attribute",
        }
    )

    level_type: str = field(
        metadata={
            "name": "LevelType",
            "type": "Attribute",
        }
    )


@dataclass
class GetMethodType:
    """Omnisensor method information."""

    formula: Formula = field(
        metadata={
            "name": "Formula",
        }
    )

    method: str = field(
        metadata={
            "name": "Method",
        }
    )

    method_hw: str = field(
        metadata={
            "name": "MethodHW",
        }
    )


@dataclass
class OmniSensor(ChildObject, Sensor):
    """OmniSensor object."""

    get: GetMethodType = field(
        metadata={
            "name": "Get",
        }
    )

    level: Union[int, Decimal, None] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def is_current_sensor(self) -> bool:
        """Return True if the sensor is a current sensor."""
        return self.model == "Current"

    @property
    def is_power_sensor(self) -> bool:
        """Return True if the sensor is a power sensor."""
        return self.model == "Power"

    @property
    def is_temperature_sensor(self) -> bool:
        """Return True if the sensor is a temperature sensor."""
        return self.model == "Temperature"



"""Controller holding and managing Vantage light sensors."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import LightSensorInterface
from aiovantage.models import LightSensor

from .base import BaseController, State


class LightSensorsController(BaseController[LightSensor], LightSensorInterface):
    """Controller holding and managing Vantage light sensors."""

    vantage_types = ("LightSensor",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("LightSensor.GetLevel",)
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a light sensor."""
        return {
            "level": await LightSensorInterface.get_level(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a light sensor."""
        if status != "LightSensor.GetLevel":
            return None

        return {
            "level": LightSensorInterface.parse_get_level_status(args),
        }



"""Keypad Station."""

from dataclasses import dataclass

from .child_object import ChildObject
from .station_object import StationObject


@dataclass
class Keypad(ChildObject, StationObject):
    """Keypad Station."""



"""Interface for querying and controlling RGB loads."""

import struct
from typing import Sequence, Tuple

from .base import Interface


class RGBLoadInterface(Interface):
    """Interface for querying and controlling RGB loads."""

    async def get_rgb(self, vid: int) -> Tuple[int, ...]:
        """Get the RGB color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the RGB color as a tuple of (red, green, blue).
        """
        return (await self.get_color(vid))[:3]

    async def get_rgbw(self, vid: int) -> Tuple[int, ...]:
        """Get the RGBW color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the RGBW color as a tuple of (red, green, blue, white).
        """
        return tuple(
            [await self.get_rgbw_channel(vid, channel) for channel in range(4)]
        )

    async def get_hsl(self, vid: int) -> Tuple[int, ...]:
        """Get the HSL color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the HSL color as a tuple of (hue, saturation, lightness).
        """
        return tuple(
            [await self.get_hsl_attribute(vid, attribute) for attribute in range(3)]
        )

    async def get_color(self, vid: int) -> Tuple[int, ...]:
        """Get the RGB/RGBW color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the RGB/RGBW color as a bytearray.
        """
        # INVOKE <id> RGBLoad.GetColor
        # -> R:INVOKE <id> <color> RGBLoad.GetColor
        response = await self.invoke(vid, "RGBLoad.GetColor")
        color = int(response.args[1])

        return tuple(struct.pack(">i", color))

    async def get_rgb_channel(self, vid: int, channel: int) -> int:
        """Get a single RGB color channel of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to get the color of.

        Returns:
            The value of the RGB channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGB <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGB <channel>
        response = await self.invoke(vid, "RGBLoad.GetRGB", channel)
        color = int(response.args[1])

        return color

    async def get_rgbw_channel(self, vid: int, channel: int) -> int:
        """Get a single RGBW color channel of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to get the color of.

        Returns:
            The value of the RGBW channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGB <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGB <channel>
        response = await self.invoke(vid, "RGBLoad.GetRGBW", channel)
        color = int(response.args[1])

        return color

    async def get_hsl_attribute(self, vid: int, attribute: int) -> int:
        """Get a single HSL color attribute of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            attribute: The attribute to get the value of.

        Returns:
            The value of the HSL attribute, 0-360 for hue, 0-100 for saturation and
            lightness.
        """
        # INVOKE <id> RGBLoad.GetHSL <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetHSL <channel>
        response = await self.invoke(vid, "RGBLoad.GetHSL", attribute)
        color = int(response.args[1])

        return color

    async def set_rgb(self, vid: int, red: float, green: float, blue: float) -> None:
        """Set the color of an RGB load.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
        """
        # Clamp levels to 0-255, ensure they're integers
        red = int(max(min(red, 255), 0))
        green = int(max(min(green, 255), 0))
        blue = int(max(min(blue, 255), 0))

        # INVOKE <id> RGBLoad.SetRGB <red> <green> <blue>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGB <red> <green> <blue>
        await self.invoke(vid, "RGBLoad.SetRGB", red, green, blue)

    async def set_rgbw(
        self, vid: int, red: float, green: float, blue: float, white: float
    ) -> None:
        """Set the color of an RGBW load.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
            white: The white value of the color, (0-255)
        """
        # Clamp levels to 0-255
        red = int(max(min(red, 255), 0))
        green = int(max(min(green, 255), 0))
        blue = int(max(min(blue, 255), 0))
        white = int(max(min(white, 255), 0))

        # INVOKE <id> RGBLoad.SetRGBW <red> <green> <blue> <white>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBW <red> <green> <blue> <white>
        await self.invoke(vid, "RGBLoad.SetRGBW", red, green, blue, white)

    async def set_hsl(
        self, vid: int, hue: float, saturation: float, lightness: float
    ) -> None:
        """Set the color of an HSL load.

        Args:
            vid: The Vantage ID of the RGB load.
            hue: The hue value of the color, in degrees (0-360).
            saturation: The saturation value of the color, in percent (0-100).
            lightness: The lightness value of the color, in percent (0-100).
        """
        # Clamp levels to 0-360, 0-100, ensure they're integers
        hue = int(max(min(hue, 360), 0))
        saturation = int(max(min(saturation, 100), 0))
        lightness = int(max(min(lightness, 100), 0))

        # INVOKE <id> RGBLoad.SetHSL <hue> <saturation> <lightness>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetHSL <hue> <saturation> <lightness>
        await self.invoke(vid, "RGBLoad.SetHSL", hue, saturation, lightness)

    async def dissolve_rgb(
        self, vid: int, red: float, green: float, blue: float, seconds: float
    ) -> None:
        """Transition the color of an RGB load over a number of seconds.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The new red value of the color, (0-255)
            green: The new green value of the color, (0-255)
            blue: The new blue value of the color, (0-255)
            seconds: The number of seconds the transition should take.
        """
        # Clamp levels to 0-255, ensure they're integers
        red = int(max(min(red, 255), 0))
        green = int(max(min(green, 255), 0))
        blue = int(max(min(blue, 255), 0))

        # INVOKE <id> RGBLoad.DissolveRGB <red> <green> <blue>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveRGB <red> <green> <blue>
        await self.invoke(vid, "RGBLoad.DissolveRGB", red, green, blue, seconds)

    async def set_rgb_component(self, vid: int, channel: int, value: float) -> None:
        """Set a single RGB(W) color channel of a load.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to set the color of.
            value: The value to set the channel to, 0-255.
        """
        # Clamp levels to 0-255, ensure they're integers
        value = int(max(min(value, 255), 0))

        # INVOKE <id> RGBLoad.SetRGBComponent <channel> <value>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBComponent <channel> <value>
        await self.invoke(vid, "RGBLoad.SetRGBComponent", channel, value)

    async def dissolve_hsl(
        self, vid: int, hue: float, saturation: float, lightness: float, seconds: float
    ) -> None:
        """Transition the color of an HSL load over a number of seconds.

        Args:
            vid: The Vantage ID of the RGB load.
            hue: The new hue value of the color, in degrees (0-360).
            saturation: The new saturation value of the color, in percent (0-100).
            lightness: The new lightness value of the color, in percent (0-100).
            seconds: The number of seconds the transition should take.
        """
        # Clamp levels to 0-360, 0-100, ensure they're integers
        hue = int(max(min(hue, 360), 0))
        saturation = int(max(min(saturation, 100), 0))
        lightness = int(max(min(lightness, 100), 0))

        # INVOKE <id> RGBLoad.DissolveHSL <hue> <saturation> <lightness>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveHSL <hue> <saturation> <lightness>
        await self.invoke(
            vid, "RGBLoad.DissolveHSL", hue, saturation, lightness, seconds
        )

    @classmethod
    def parse_color_channel_status(cls, args: Sequence[str]) -> Tuple[int, int]:
        """Parse an 'RGBLoad.GetRGB' event."""
        # ELLOG STATUSEX ON
        # -> EL: <id> RGBLoad.GetRGB <value> <channel>
        # STATUS ADD <id>
        # -> S:STATUS <id> RGBLoad.GetRGB <value> <channel>
        value = int(args[0])
        channel = int(args[1])

        return channel, value

    @classmethod
    def parse_get_color_status(cls, args: Sequence[str]) -> bytearray:
        """Parse an 'RGBLoad.GetColor' event."""
        # ELLOG STATUS ON
        # -> EL: <id> RGBLoad.GetColor <color>
        # STATUS ADD <id>
        # -> S:STATUS <id> RGBLoad.GetColor <color>
        color = int(args[0])

        return bytearray(struct.pack(">i", color))



"""Controller holding and managing Vantage tasks."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import TaskInterface
from aiovantage.models import Task

from .base import BaseController, State


class TasksController(BaseController[Task], TaskInterface):
    """Controller holding and managing Vantage tasks."""

    vantage_types = ("Task",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("Task.IsRunning", "Task.GetState")
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a task."""
        return {
            "is_running": await TaskInterface.is_running(self, vid),
            "state": await TaskInterface.get_state(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a task."""
        if status == "Task.IsRunning":
            return {
                "is_running": TaskInterface.parse_is_running_status(args),
            }

        if status == "Task.GetState":
            return {
                "state": TaskInterface.parse_get_state_status(args),
            }

        return None



"""Use a CommandClient to subscribe to status updates for LOAD objects."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage.command_client import Event, EventStream, EventType

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def command_client_callback(event: Event) -> None:
    """Print out the status update for each event."""
    if event["type"] == EventType.STATUS:
        print(f"[{event['status_type']}] id: {event['id']}, args: {event['args']}")
    elif event["type"] == EventType.CONNECTED:
        print("Connected and monitoring for status updates...")
    elif event["type"] == EventType.DISCONNECTED:
        print("Disconnected")
    elif event["type"] == EventType.RECONNECTED:
        print("Reconnected")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Create an EventStream client
    async with EventStream(args.host, args.username, args.password) as events:
        # Subscribe to connection events
        events.subscribe(
            command_client_callback,
            (EventType.CONNECTED, EventType.DISCONNECTED, EventType.RECONNECTED),
        )

        # Subscribe to status updates for LOAD objects (STATUS LOAD)
        events.subscribe_status(command_client_callback, "LOAD")

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Interface for querying and controlling variables."""

from typing import Sequence, Union

from .base import Interface


class GMemInterface(Interface):
    """Interface for querying and controlling variables."""

    async def get_value(self, vid: int) -> str:
        """Get the value of a variable.

        Args:
            vid: The Vantage ID of the variable.

        Returns:
            The value of the variable, as a raw string.
        """
        # GETVARIABLE {id}
        # -> R:GETVARIABLE {id} {value}
        response = await self.command_client.command("GETVARIABLE", vid)
        value = response.args[1]

        return value

    async def set_value(self, vid: int, value: Union[int, str, bool]) -> None:
        """Set the value of a variable.

        Args:
            vid: The Vantage ID of the variable.
            value: The value to set, either a bool, int, or str.
        """
        # SETVARIABLE {id} {value}
        # -> R:SETVARIABLE {id} {value}
        await self.command_client.command("VARIABLE", vid, value, force_quotes=True)

    @classmethod
    def parse_variable_status(cls, args: Sequence[str]) -> str:
        """Parse a simple 'S:VARIABLE' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the load.
        """
        # STATUS VARIABLE
        # -> S:VARIABLE <id> <value>
        return args[0]



"""Base controller for Vantage objects."""

import asyncio
import logging
from dataclasses import fields
from inspect import iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from aiovantage.command_client import CommandClient, Event, EventStream, EventType
from aiovantage.command_client.utils import tokenize_response
from aiovantage.config_client import ConfigClient
from aiovantage.config_client.requests import get_objects
from aiovantage.events import EventCallback, VantageEvent
from aiovantage.models import SystemObject
from aiovantage.query import QuerySet

if TYPE_CHECKING:
    from aiovantage import Vantage

T = TypeVar("T", bound=SystemObject)


# Types for state and subscriptions
EventSubscription = Tuple[EventCallback[T], Optional[Iterable[VantageEvent]]]
State = Optional[Dict[str, Any]]


class BaseController(QuerySet[T]):
    """Base controller for Vantage objects."""

    vantage_types: Tuple[str, ...]
    """The Vantage object types that this controller handles."""

    status_types: Optional[Tuple[str, ...]] = None
    """Which Vantage 'STATUS' types this controller handles, if any."""

    enhanced_log_status_methods: Optional[Union[Tuple[str, ...], Literal["*"]]] = None
    """Which status methods this controller handles from the Enhanced Log."""

    def __init__(self, vantage: "Vantage") -> None:
        """Initialize a controller.

        Args:
            vantage: The Vantage instance.
            init_callback: A callback to call when the controller is initialized.
        """
        self._vantage = vantage
        self._items: Dict[int, T] = {}
        self._logger = logging.getLogger(__package__)
        self._subscribed_to_state_changes = False
        self._subscriptions: List[EventSubscription[T]] = []
        self._id_subscriptions: Dict[int, List[EventSubscription[T]]] = {}
        self._initialized = False

        QuerySet.__init__(self, self._items, self._lazy_initialize)

        self.__post_init__()

    def __post_init__(self) -> None:
        """Post initialization hook for subclasses."""

    def __getitem__(self, vid: int) -> T:
        """Return the object with the given Vantage ID."""
        return self._items[vid]

    def __contains__(self, vid: int) -> bool:
        """Return True if the object with the given Vantage ID exists."""
        return vid in self._items

    @property
    def config_client(self) -> ConfigClient:
        """Return the config client instance."""
        return self._vantage.config_client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client instance."""
        return self._vantage.command_client

    @property
    def event_stream(self) -> EventStream:
        """Return the event stream instance."""
        return self._vantage.event_stream

    @property
    def initialized(self) -> bool:
        """Return True if this controller has been initialized."""
        return self._initialized

    @property
    def stateful(self) -> bool:
        """Return True if this controller manages stateful objects."""
        return bool(self.status_types or self.enhanced_log_status_methods)

    @property
    def known_ids(self) -> Set[int]:
        """Return a set of all known object IDs."""
        return set(self._items.keys())

    async def fetch_object_state(self, _vid: int) -> State:
        """Fetch the full state of an object, should be overridden by subclasses."""
        return None

    def parse_object_update(
        self, _vid: int, _status: str, _args: Sequence[str]
    ) -> State:
        """Parse updates from the event stream for an object, should be overridden by subclasses."""
        return None

    async def initialize(self, fetch_state: bool = True) -> None:
        """Populate objects and fetch their initial state."""
        prev_ids = set(self._items.keys())
        cur_ids = set()

        # Fetch all objects managed by this controller
        async for obj in get_objects(self.config_client, types=self.vantage_types):
            if obj.id not in prev_ids:
                # This is a new object, add it to the controller
                self._items[obj.id] = obj

                # Fetch the state of the object
                if self.stateful and fetch_state:
                    self._set_state(obj, await self.fetch_object_state(obj.id))

                # Notify subscribers that a new object was added
                self.emit(VantageEvent.OBJECT_ADDED, obj)
            else:
                # This is an existing object, check if any attributes have changed
                prev_obj = self._items[obj.id]
                attrs_changed = [
                    field.name
                    for field in fields(prev_obj)
                    if getattr(prev_obj, field.name) != getattr(obj, field.name)
                    and field.name != "mtime"
                ]

                # If any attributes changed, update the object and notify subscribers
                if attrs_changed:
                    for attr in attrs_changed:
                        try:
                            setattr(prev_obj, attr, getattr(obj, attr))
                        except AttributeError:
                            self._logger.warning("Object has no attribute '%s'", attr)

                    self.emit(
                        VantageEvent.OBJECT_UPDATED,
                        obj,
                        {"attrs_changed": attrs_changed},
                    )

            # Keep track of which objects we've seen
            cur_ids.add(obj.id)

        # Handle objects that were removed
        for vid in prev_ids - cur_ids:
            obj = self._items.pop(vid)
            self.emit(VantageEvent.OBJECT_DELETED, obj)

        # Subscribe to state changes for objects managed by this controller
        if fetch_state:
            await self.subscribe_to_state_changes()

        # Mark the controller as initialized
        if not self._initialized:
            self._initialized = True
            self._logger.info("%s initialized", self.__class__.__name__)
        else:
            self._logger.info("%s reinitialized", self.__class__.__name__)

    async def fetch_full_state(self) -> None:
        """Fetch the full state of all objects managed by this controller."""
        if not self.stateful:
            return

        for obj in self._items.values():
            self._update_state(obj.id, await self.fetch_object_state(obj.id))

        self._logger.info("%s fetched state", self.__class__.__name__)

    async def subscribe_to_state_changes(self) -> None:
        """Subscribe to state changes for objects managed by this controller."""
        if self._subscribed_to_state_changes or not self.stateful:
            return

        # Ensure that the event stream is running
        await self.event_stream.start()

        # Subscribe to "STATUS {type}" updates, if this controller cares about them
        if self.status_types:
            self.event_stream.subscribe_status(self._handle_event, self.status_types)

        # Subscribe to object status events from the "Enhanced Log"
        if self.enhanced_log_status_methods:
            self.event_stream.subscribe_enhanced_log(
                self._handle_event, ("STATUS", "STATUSEX")
            )

        self._subscribed_to_state_changes = True
        self._logger.info("%s subscribed to state changes", self.__class__.__name__)

    def subscribe(
        self,
        callback: EventCallback[T],
        id_filter: Union[int, Tuple[int], None] = None,
        event_filter: Union[VantageEvent, Tuple[VantageEvent, ...], None] = None,
    ) -> Callable[[], None]:
        """Subscribe to status changes for objects managed by this controller.

        Args:
            callback: The callback to call when an object changes.
            id_filter: The Vantage IDs to subscribe to, all objects if None.
            event_filter: The event types to subscribe to, all events if None.

        Returns:
            A function to unsubscribe from the callback.
        """
        # Handle single ID filter or single event filter
        if isinstance(id_filter, int):
            id_filter = (id_filter,)

        if isinstance(event_filter, VantageEvent):
            event_filter = (event_filter,)

        # Create the subscription
        subscription = (callback, event_filter)

        # Add the subscription to the list of subscriptions
        if id_filter is None:
            self._subscriptions.append(subscription)
        else:
            for vid in id_filter:
                if vid not in self._id_subscriptions:
                    self._id_subscriptions[vid] = []
                self._id_subscriptions[vid].append(subscription)

        # Return a function to unsubscribe
        def unsubscribe() -> None:
            if id_filter is None:
                self._subscriptions.remove(subscription)
            else:
                for vid in id_filter:
                    if vid not in self._id_subscriptions:
                        continue
                    self._id_subscriptions[vid].remove(subscription)

        return unsubscribe

    def emit(
        self, event_type: VantageEvent, obj: T, data: Optional[Dict[str, Any]] = None
    ) -> None:
        """Emit an event to subscribers of this controller.

        Args:
            event_type: The type of event to emit.
            obj: The object that the event relates to.
            data: Data to pass to the callback.
        """
        if data is None:
            data = {}

        # Grab a list of subscribers that care about this object
        subscribers = self._subscriptions + self._id_subscriptions.get(obj.id, [])
        for callback, event_filter in subscribers:
            if event_filter is not None and event_type not in event_filter:
                continue

            if iscoroutinefunction(callback):
                asyncio.create_task(callback(event_type, obj, data))  # noqa: RUF006
            else:
                callback(event_type, obj, data)

    def _set_state(self, obj: T, state: State) -> None:
        # Set the state properties of an object.
        if state is None:
            return

        for key, value in state.items():
            try:
                setattr(obj, key, value)
            except AttributeError:
                self._logger.warning("Object '%d' has no attribute '%s'", obj.id, key)

    def _update_state(self, vid: int, state: State) -> None:
        """Update the state of an object and notify subscribers if it changed."""
        if state is None:
            return

        # Ignore updates for objects that this controller doesn't manage
        if (obj := self._items.get(vid)) is None:
            return

        # Check if any of the attributes changed and update them
        attrs_changed = []
        for key, value in state.items():
            try:
                if getattr(obj, key) != value:
                    setattr(obj, key, value)
                    attrs_changed.append(key)
            except AttributeError:
                self._logger.warning("Object '%d' has no attribute '%s'", obj.id, key)

        # Notify subscribers if any attributes changed
        if len(attrs_changed) > 0:
            self.emit(
                VantageEvent.OBJECT_UPDATED,
                obj,
                {"attrs_changed": attrs_changed},
            )

    async def _handle_event(self, event: Event) -> None:
        # Handle events from the event stream
        # pylint: disable=assignment-from-none
        if event["type"] == EventType.STATUS:
            # Ignore events for objects that this controller doesn't manage
            if event["id"] not in self._items:
                return

            # Pass the event to the controller
            state = self.parse_object_update(
                event["id"], event["status_type"], event["args"]
            )
            self._update_state(event["id"], state)

        elif event["type"] == EventType.ENHANCED_LOG:
            # STATUS/STATUSEX logs can be tokenized the same as command responses
            id_str, method, *args = tokenize_response(event["log"])

            # Ignore events with methods that this controller doesn't care about
            if not (
                self.enhanced_log_status_methods
                and (
                    self.enhanced_log_status_methods == "*"
                    or method in self.enhanced_log_status_methods
                )
            ):
                return

            # Ignore events for objects that this controller doesn't manage
            vid = int(id_str)
            if vid not in self._items:
                return

            # Pass the event to the controller
            state = self.parse_object_update(vid, method, args)
            self._update_state(vid, state)

    async def _lazy_initialize(self) -> None:
        # Initialize the controller if it isn't already initialized
        if not self._initialized:
            await self.initialize()



"""Vantage events."""

from enum import Enum
from typing import Any, Callable, Dict, TypeAlias, TypeVar

from .models import SystemObject


class VantageEvent(Enum):
    """Enum with possible Events."""

    OBJECT_ADDED = "add"
    OBJECT_UPDATED = "update"
    OBJECT_DELETED = "delete"


T = TypeVar("T", bound=SystemObject)

EventCallback: TypeAlias = Callable[[VantageEvent, T, Dict[str, Any]], None]
"""Type alias for a Vantage event callback function."""



"""IIntrospection.GetVersion method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional


@dataclass
class GetVersion:
    """IIntrospection.GetVersion method definition."""

    interface: ClassVar[str] = "IIntrospection"
    call = None
    return_value: Optional["GetVersion.Return"] = field(
        default=None,
        metadata={
            "name": "return",
        },
    )

    @dataclass
    class Return:
        """IIntrospection.GetVersion method return value."""

        kernel: Optional[str] = field(default=None)
        rootfs: Optional[str] = field(default=None)
        app: Optional[str] = field(default=None)



"""Station bus object."""

from dataclasses import dataclass

from .child_object import ChildObject
from .system_object import SystemObject


@dataclass
class StationBus(ChildObject, SystemObject):
    """Station bus object."""



"""Provides a basic "Django-ish" queryset for querying objects."""

from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    overload,
)

T = TypeVar("T")


class QuerySet(Iterable[T], AsyncIterator[T]):
    """Queryset class for querying objects from a dictionary.

    Querysets are iterable and async iterable, and can be chained together to
    filter objects.
    """

    def __init__(
        self,
        data: Dict[int, T],
        populate: Callable[[], Awaitable[None]],
        filters: Optional[List[Callable[[T], Any]]] = None,
    ) -> None:
        """Initialize a queryset.

        Args:
            data: The data dictionary to query.
            populate: A coroutine to populate the data so we don't have a complete
                      dataset before using "async for" loops.
            filters: A list of filters to apply to the queryset.
        """
        self.__data = data
        self.__populate = populate
        self.__iterator: Optional[Iterator[T]] = None

        if filters is None:
            self.__filters: List[Callable[[T], Any]] = []
        else:
            self.__filters = filters

    def __iter__(self) -> Iterator[T]:
        """Return an iterator over the queryset."""
        for obj in self.__data.values():
            if all(filter_fn(obj) for filter_fn in self.__filters):
                yield obj

    def __bool__(self) -> bool:
        """Return True if the queryset contains any objects."""
        return any(True for _ in self)

    def __aiter__(self) -> AsyncIterator[T]:
        """Return an async iterator over the queryset."""
        self.__iterator = None
        return self

    async def __anext__(self) -> T:
        """Return the next object in the queryset."""
        if self.__iterator is None:
            if self.__populate is not None:
                await self.__populate()

            self.__iterator = iter(self)

        try:
            return next(self.__iterator)
        except StopIteration as exc:
            raise StopAsyncIteration from exc

    def add_filter(self, filter_fn: Callable[[T], Any]) -> None:
        """Add a filter to the queryset."""
        self.__filters.append(filter_fn)

    @overload
    def filter(self, match: Callable[[T], Any]) -> "QuerySet[T]":
        ...

    @overload
    def filter(self, **kwargs: Any) -> "QuerySet[T]":
        ...

    def filter(self, *args: Any, **kwargs: Any) -> "QuerySet[T]":
        """Return a queryset of items that match the given filter."""
        queryset = QuerySet(self.__data, self.__populate, self.__filters.copy())

        if len(args) == 1:
            queryset.add_filter(args[0])
        elif len(args) == 0 and len(kwargs) > 0:
            queryset.add_filter(
                lambda obj: all(
                    getattr(obj, key) == value for key, value in kwargs.items()
                )
            )
        else:
            raise TypeError("filter() and get() expect either a callable or **kwargs")

        return queryset

    @overload
    def get(self, key: int, default: Optional[T] = None) -> Optional[T]:
        ...

    @overload
    def get(self, match: Callable[[T], Any]) -> Optional[T]:
        ...

    @overload
    def get(self, **kwargs: Any) -> Optional[T]:
        ...

    def get(self, *args: Any, **kwargs: Any) -> Optional[T]:
        """Get the first object that matches the given filter."""
        # Handle the case where we're getting an object by key
        if len(args) == 1 and isinstance(args[0], int):
            return self.__data.get(args[0], kwargs.get("default", None))

        # Otherwise, pass through to filter and return the first object
        return next(iter(self.filter(*args, **kwargs)), None)

    @overload
    async def aget(self, key: int, default: Optional[T] = None) -> Optional[T]:
        ...

    @overload
    async def aget(self, match: Callable[[T], Any]) -> Optional[T]:
        ...

    @overload
    async def aget(self, **kwargs: Any) -> Optional[T]:
        ...

    async def aget(self, *args: Any, **kwargs: Any) -> Optional[T]:
        """Asynchronously get the first object that matches the given filter."""
        await self.__populate()
        return self.get(*args, **kwargs)

    def first(self) -> Optional[T]:
        """Return the first object in the queryset."""
        return next(iter(self), None)

    async def afirst(self) -> Optional[T]:
        """Asynchronously return the first object in the queryset."""
        await self.__populate()
        return self.first()



"""Controller holding and managing Vantage controllers."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import IntrospectionInterface, ObjectInterface
from aiovantage.models import Master

from .base import BaseController, State


class MastersController(
    BaseController[Master],
    ObjectInterface,
    IntrospectionInterface,
):
    """Controller holding and managing Vantage controllers."""

    vantage_types = ("Master",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("Object.GetMTime",)
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a Vantage controller."""
        return {
            "last_updated": await ObjectInterface.get_mtime(self, vid),
            "firmware_version": await IntrospectionInterface.get_firmware_version(
                self, vid, IntrospectionInterface.Firmware.APPLICATION
            ),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a Vantage controller."""
        if status != "Object.GetMTime":
            return None

        return {
            "last_updated": ObjectInterface.parse_get_mtime_status(args),
        }



"""Interface for querying and controlling tasks."""

from typing import Sequence

from .base import Interface


class TaskInterface(Interface):
    """Interface for querying and controlling tasks."""

    async def is_running(self, vid: int) -> bool:
        """Get the running state of a task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.IsRunning
        # -> R:INVOKE <id> <running (0/1)> Task.IsRunning
        response = await self.invoke(vid, "Task.IsRunning")
        is_running = bool(int(response.args[1]))

        return is_running

    async def get_state(self, vid: int) -> bool:
        """Get the state of a task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.GetState
        # -> R:INVOKE <id> <state (0/1)> Task.GetState
        response = await self.invoke(vid, "Task.GetState")
        task_state = bool(int(response.args[1]))

        return task_state

    async def start(self, vid: int) -> None:
        """Start a task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.Start <source> <event> <param1> <param2>
        # -> R:INVOKE <id> <rcode (0/1)> Task.Start <source> <event> <param1> <param2>
        await self.invoke(vid, "Task.Start")

    async def stop(self, vid: int) -> None:
        """Stop a running task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.Stop
        # -> R:INVOKE <id> <rcode> Task.Stop
        await self.invoke(vid, "Task.Stop")

    @classmethod
    def parse_task_status(cls, args: Sequence[str]) -> bool:
        """Parse a simple 'S:TASK' event.

        Args:
            args: The arguments of the event.

        Returns:
            The state of the task.
        """
        # STATUS TASK
        # -> S:TASK <id> <state (0/1)>
        return bool(int(args[0]))

    @classmethod
    def parse_get_state_status(cls, args: Sequence[str]) -> bool:
        """Parse a 'Task.GetState' event.

        Args:
            args: The arguments of the event.

        Returns:
            The state of the task.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Task.GetState <state (0/1)>
        # STATUS ADD <id>
        # -> S:STATUS <id> Task.GetState <state (0/1)>
        return bool(int(args[0]))

    @classmethod
    def parse_is_running_status(cls, args: Sequence[str]) -> bool:
        """Parse a 'Task.IsRunning' event.

        Args:
            args: The arguments of the event.

        Returns:
            The running state of the task.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Task.IsRunning <running (0/1)>
        # STATUS ADD <id>
        # -> S:STATUS <id> Task.IsRunning <running (0/1)>
        return bool(int(args[0]))



"""Qube blind object."""

from dataclasses import dataclass

from .blind_base import BlindBase
from .station_object import StationObject


@dataclass
class QubeBlind(BlindBase, StationObject):
    """Qube blind object."""



"""DMX/DALI Gateway station."""

from dataclasses import dataclass

from .station_object import StationObject


@dataclass
class VantageDmxDaliGateway(StationObject):
    """DMX/DALI Gateway station."""

    class Meta:
        name = "Vantage.DmxDaliGateway"



"""Interface for querying and controlling color temperature."""

from typing import Sequence

from .base import Interface


class ColorTemperatureInterface(Interface):
    """Interface for querying and controlling color temperature."""

    async def get_color_temp(self, vid: int) -> int:
        """Get the color temperature of a light.

        Args:
            vid: The Vantage ID of the light.

        Returns:
            The color temperature of the light, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.Get
        # -> R:INVOKE <id> <temp> ColorTemperature.Get
        response = await self.invoke(vid, "ColorTemperature.Get")
        color_temp = int(response.args[1])

        return color_temp

    async def set_color_temp(self, vid: int, temp: int, transition: int = 0) -> None:
        """Set the color temperature of a light.

        Args:
            vid: The Vantage ID of the light.
            temp: The color temperature to set the light to, in Kelvin.
            transition: The time in seconds to transition to the new color
        """
        # Ensure the temperature is an integer
        temp = int(temp)

        # INVOKE <id> ColorTemperature.Set <temp> <seconds>
        # -> R:INVOKE <id> <rcode> ColorTemperature.Set <temp>
        await self.invoke(vid, "ColorTemperature.Set", temp, transition)

    @classmethod
    def parse_get_status(cls, args: Sequence[str]) -> int:
        """Parse a 'ColorTemperature.Get' event.

        Args:
            args: The arguments of the event.

        Returns:
            The color temperature, in Kelvin.
        """
        # ELLOG STATUS ON
        # -> EL: <id> ColorTemperature.Get <temp>
        # STATUS ADD <id>
        # -> S:STATUS <id> ColorTemperature.Get <temp>
        color_temp = int(args[0])

        return color_temp



"""Controller holding and managing Vantage blinds."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import BlindInterface
from aiovantage.models import BlindBase

from .base import BaseController, State


class BlindsController(BaseController[BlindBase], BlindInterface):
    """Controller holding and managing Vantage blinds."""

    vantage_types = (
        "QISBlind",
        "QubeBlind",
        "RelayBlind",
        "Somfy.RS-485_Shade_CHILD",
        "Somfy.URTSI_2_Shade_CHILD",
    )
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("Blind.GetPosition",)
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a blind."""
        return {
            "position": await BlindInterface.get_position(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a blind."""
        if status != "Blind.GetPosition":
            return None

        return {
            "position": BlindInterface.parse_get_position_status(args),
        }



"""Controller holding and managing Vantage modules."""

from aiovantage.models import SystemObject

from .base import BaseController


class ModulesController(BaseController[SystemObject]):
    """Controller holding and managing Vantage power modules."""

    vantage_types = ("Module", "ModuleGen2")
    """The Vantage object types that this controller will fetch."""



"""DMX/DALI Gateway color load object."""

from dataclasses import dataclass

from .rgb_load_base import RGBLoadBase


@dataclass
class VantageDDGColorLoad(RGBLoadBase):
    """DMX/DALI Gateway color load object."""

    class Meta:
        name = "Vantage.DDGColorLoad"



"""Interface for querying and controlling loads."""

from enum import IntEnum
from typing import Sequence

from .base import Interface


class LoadInterface(Interface):
    """Interface for querying and controlling loads."""

    class RampType(IntEnum):
        """The type of ramp to perform."""

        STOP = 2
        OPPOSITE = 3
        DOWN = 4
        UP = 5
        FIXED = 6
        VARIABLE = 7
        ADJUST = 8

    async def turn_on(
        self, vid: int, transition: float = 0, level: float = 100
    ) -> None:
        """Turn on a load.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to transition to the new level.
            level: The level to set the load to (0-100).
        """
        if transition:
            await self.ramp(vid, level, transition)
        else:
            await self.set_level(vid, level)

    async def turn_off(self, vid: int, transition: float = 0) -> None:
        """Turn off a load.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to ramp the load down.
        """
        if transition:
            await self.ramp(vid, 0, transition)
        else:
            await self.set_level(vid, 0)

    async def get_level(self, vid: int) -> float:
        """Get the level of a load.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetLevel
        # -> R:INVOKE <id> <level (0-100)> Load.GetLevel
        response = await self.invoke(vid, "Load.GetLevel")
        level = float(response.args[1])

        return level

    async def set_level(self, vid: int, level: float) -> None:
        """Set the level of a load.

        Args:
            vid: The Vantage ID of the load.
            level: The level to set the load to (0-100).
        """
        # Clamp level to 0-100
        level = max(min(level, 100), 0)

        # INVOKE <id> Load.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Load.SetLevel <level (0-100)>
        await self.invoke(vid, "Load.SetLevel", level)

    async def ramp(
        self,
        vid: int,
        level: float,
        seconds: float,
        ramp_type: RampType = RampType.FIXED,
    ) -> None:
        """Ramp a load to a level over a number of seconds.

        Args:
            vid: The Vantage ID of the load.
            level: The level to ramp the load to (0-100).
            seconds: The number of seconds to ramp the load over.
            ramp_type: The type of ramp to perform.
        """
        # INVOKE <id> Load.Ramp <type> <seconds> <level>
        # -> R:INVOKE <id> <rcode> Load.Ramp <type> <seconds> <level>
        await self.invoke(vid, "Load.Ramp", ramp_type, seconds, level)

    @classmethod
    def parse_load_status(cls, args: Sequence[str]) -> float:
        """Parse a simple 'S:LOAD' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the load.
        """
        # STATUS LOAD
        # -> S:LOAD <id> <level (0-100)>
        return float(args[0])

    @classmethod
    def parse_get_level_status(cls, args: Sequence[str]) -> float:
        """Parse a 'Load.GetLevel' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the load.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Load.GetLevel <level (0-100000)>
        # ADDSTATUS <id>
        # -> S:STATUS <id> Load.GetLevel <level (0-100000)>
        return float(args[0]) / 1000



"""LoadGroup object."""

from dataclasses import dataclass, field
from typing import List, Optional

from .location_object import LocationObject


@dataclass
class LoadGroup(LocationObject):
    """LoadGroup object."""

    load_ids: List[int] = field(
        default_factory=list,
        metadata={
            "name": "Load",
            "wrapper": "LoadTable",
        },
    )

    level: Optional[float] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



"""Interface for querying and controlling sensors."""

from typing import Sequence

from .base import Interface


class SensorInterface(Interface):
    """Interface for querying and controlling sensors."""

    async def get_level(self, vid: int) -> int:
        """Get the level of a sensor.

        Args:
            vid: The Vantage ID of the sensor.
        """
        # INVOKE <id> Sensor.GetLevel
        # -> R:INVOKE <id> <level (0-100)> Sensor.GetLevel
        response = await self.invoke(vid, "Sensor.GetLevel")
        level = int(response.args[1])

        return level

    @classmethod
    def parse_get_level_status(cls, args: Sequence[str]) -> int:
        """Parse a 'Sensor.GetLevel' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the sensor.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Sensor.GetLevel <level>
        # STATUS ADD <id>
        # -> S:STATUS <id> Sensor.GetLevel <level>
        return int(args[0])



"""Controller holding and managing Vantage areas."""

from aiovantage.models import Area

from .base import BaseController


class AreasController(BaseController[Area]):
    """Controller holding and managing Vantage areas."""

    vantage_types = ("Area",)
    """The Vantage object types that this controller will fetch."""



"""Power profile object."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass
class PowerProfile(SystemObject):
    """Power Profile object."""

    min: float = field(
        metadata={
            "name": "Min",
        }
    )

    max: float = field(
        metadata={
            "name": "Max",
        }
    )

    adjust: int = field(
        metadata={
            "name": "Adjust",
        }
    )

    freq: int = field(
        metadata={
            "name": "Freq",
        }
    )

    inductive: bool = field(
        metadata={
            "name": "Inductive",
        }
    )



"""Controller holding and managing Vantage omni sensors."""

from decimal import Decimal
from typing import Sequence, Union

from typing_extensions import override

from aiovantage.models import OmniSensor

from .base import BaseController, State


class OmniSensorsController(BaseController[OmniSensor]):
    """Controller holding and managing Vantage omni sensors."""

    vantage_types = ("OmniSensor",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = "*"
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of an omni sensor."""
        return {
            "level": await self.get_level(vid),
        }

    @override
    def parse_object_update(self, vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for an omni sensor."""
        omni_sensor = self[vid]
        if status != omni_sensor.get.method:
            return None

        return {
            "level": self.parse_get_level_status(omni_sensor, args),
        }

    async def get_level(self, vid: int, cached: bool = False) -> Union[int, Decimal]:
        """Get the level of an OmniSensor.

        Args:
            vid: The ID of the sensor.
            cached: Whether to use the cached value or fetch a new one.

        Returns:
            The level of the sensor.
        """
        # Figure out which get method to use, hardware or software (cached)
        omni_sensor = self[vid]
        method = omni_sensor.get.method if cached else omni_sensor.get.method_hw

        # INVOKE <id> <method>
        # -> R:INVOKE <id> <value> <method>
        response = await self.command_client.command("INVOKE", vid, method)

        # Convert the level to the correct type
        if omni_sensor.get.formula.return_type == "fixed":
            level = Decimal(response.args[1])
            if omni_sensor.get.formula.level_type == "int":
                return int(level)

            return level

        if omni_sensor.get.formula.return_type == "int":
            level = Decimal(response.args[1]) / 1000
            if omni_sensor.get.formula.level_type == "int":
                return int(level)

            return level

        raise ValueError(f"Unknown return type {omni_sensor.get.formula.return_type}")

    @classmethod
    def parse_get_level_status(
        cls, omni_sensor: OmniSensor, args: Sequence[str]
    ) -> Union[int, Decimal]:
        """Parse an OmniSensor 'GetLevel' event, eg. 'PowerSensor.GetPower'."""
        # ELLOG STATUS ON
        # -> EL: <id> <method> <value>
        # STATUS ADD <id>
        # -> S:STATUS <id> <method> <value>
        level = Decimal(args[0]) / 1000
        if omni_sensor.get.formula.level_type == "int":
            return int(level)

        return level



"""Blind base class."""

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Optional

from .system_object import SystemObject


@dataclass
class BlindBase(SystemObject):
    """Blind base class."""

    orientation: Optional[str] = field(
        default=None,
        metadata={
            "name": "ShadeOrientation",
            "type": "Attribute",
        },
    )

    type: Optional[str] = field(
        default=None,
        metadata={
            "name": "ShadeType",
            "type": "Attribute",
        },
    )

    position: Optional[Decimal] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Fetch all anemo sensors from the Vantage controller, and print any state changes."""

import argparse
import asyncio
import contextlib
import logging
from typing import Any, Dict

from aiovantage import Vantage, VantageEvent
from aiovantage.models import AnemoSensor

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def callback(event: VantageEvent, obj: AnemoSensor, data: Dict[str, Any]) -> None:
    """Print out any state changes."""
    if event == VantageEvent.OBJECT_ADDED:
        print(f"[Sensor added] '{obj.name}' ({obj.id})")

    elif event == VantageEvent.OBJECT_UPDATED:
        print(f"[Sensor updated] '{obj.name}' ({obj.id})")
        for attr in data.get("attrs_changed", []):
            print(f"    {attr} = {getattr(obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Subscribe to updates for all sensors
        vantage.anemo_sensors.subscribe(callback)

        # Fetch all known sensors from the controller
        await vantage.anemo_sensors.initialize()

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Controller holding and managing Vantage dry contacts."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import ButtonInterface
from aiovantage.models import DryContact

from .base import BaseController, State


class DryContactsController(BaseController[DryContact], ButtonInterface):
    """Controller holding and managing Vantage dry contacts."""

    vantage_types = ("DryContact",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("BTN",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a dry contact."""
        if status != "BTN":
            return None

        return {
            "triggered": ButtonInterface.parse_btn_status(args),
        }



"""Controller holding and managing Vantage variables."""
import re
from typing import Sequence, Union

from typing_extensions import override

from aiovantage.command_client.interfaces import GMemInterface
from aiovantage.command_client.utils import parse_byte_param
from aiovantage.models import GMem

from .base import BaseController, State


class GMemController(BaseController[GMem], GMemInterface):
    """Controller holding and managing Vantage variables."""

    vantage_types = ("GMem",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("VARIABLE",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a variable."""
        return {
            "value": self._parse_value(vid, await self.get_value(vid)),
        }

    @override
    def parse_object_update(self, vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a variable."""
        if status != "VARIABLE":
            return None

        return {
            "value": self._parse_value(vid, GMemInterface.parse_variable_status(args)),
        }

    def _parse_value(self, vid: int, value: str) -> Union[int, str, bool]:
        # Parse the results of a GMem lookup into the expected type.
        gmem: GMem = self[vid]

        if gmem.is_bool:
            return bool(int(value))
        if gmem.is_str:
            # Handle byte array strings
            if re.match(r"^[\[\{].*[\]\}]$", value):
                byte_param = parse_byte_param(value)
                return byte_param.decode().rstrip("\x00")

            return value

        return int(value)



"""Task object."""

from dataclasses import dataclass, field
from typing import Optional

from .system_object import SystemObject


@dataclass
class Task(SystemObject):
    """Task object."""

    is_running: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    state: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Utility functions for the Host Command service client."""

import re
import struct
from decimal import Decimal, InvalidOperation
from typing import Sequence, Type, Union

TOKEN_PATTERN = re.compile(r'"([^""]*(?:""[^""]*)*)"|(\{.*?\})|(\[.*?\])|(\S+)')

ParameterType = Union[str, bool, int, float, Decimal, bytearray]


def tokenize_response(string: str) -> Sequence[str]:
    """Tokenize a response from the Host Command service.

    Handles quoted strings and byte arrays as single tokens.

    Args:
        string: The response string to tokenize.

    Returns:
        A list of string tokens.
    """
    tokens = []
    for match in TOKEN_PATTERN.finditer(string):
        token = match.group(0)

        # Remove quotes from quoted strings, and unescape quotes
        if token.startswith('"') and token.endswith('"'):
            token = token[1:-1].replace('""', '"')

        tokens.append(token)

    return tokens


def parse_params(
    params: Sequence[str], signature: Sequence[Type[ParameterType]]
) -> Sequence[ParameterType]:
    """Parse response parameters from the Host Command service.

    Handles parsing tokens into the expected types, as defined by the signature.

    Args:
        params: The parameters to parse.
        signature: The expected parameter types.

    Returns:
        A list of parameters of the correct type.

    Raises:
        ValueError: If the parameter count does not match the signature, if a
            parameter is of an unsupported type, or if a parameter cannot be parsed.
    """
    parsed_params = []
    for index, param in enumerate(params):
        if index >= len(signature):
            raise ValueError("More parameters than expected")

        parsed_param: ParameterType
        if signature[index] == str:
            parsed_param = parse_string_param(param)
        elif signature[index] == bool:
            parsed_param = bool(int(param))
        elif signature[index] == int:
            parsed_param = int(param)
        elif signature[index] == Decimal:
            try:
                parsed_param = Decimal(param)
            except InvalidOperation as err:
                raise ValueError from err
        elif signature[index] == bytearray:
            parsed_param = parse_byte_param(param)
        else:
            raise ValueError("Invalid parameter type")

        parsed_params.append(parsed_param)

    return parsed_params


def encode_params(*params: ParameterType, force_quotes: bool = False) -> str:
    """Encode a list of parameters for sending to the Host Command service.

    Converts all params to strings, wraps strings in double quotes, and escapes
    double quotes.

    Args:
        params: The parameters to encode.
        force_quotes: Whether to force string params to be wrapped in double quotes.

    Returns:
        The encoded parameters, joined by spaces.

    Raises:
        TypeError: If a parameter is of an unsupported type.
    """
    encoded_params = []
    for value in params:
        if isinstance(value, str):
            encoded_param = encode_string_param(value, force_quotes)
        elif isinstance(value, bool):
            encoded_param = "1" if value else "0"
        elif isinstance(value, (int, float, Decimal)):
            encoded_param = str(value)
        elif isinstance(value, bytearray):
            encoded_param = encode_byte_param(value)
        else:
            raise TypeError(f"Invalid value type: {type(value)}")

        encoded_params.append(encoded_param)

    return " ".join(encoded_params)


def parse_string_param(param: str) -> str:
    """Parse a string parameter from the Host Command service.

    Handles unescaping quotes.

    Args:
        param: The parameter to parse.

    Returns:
        The parsed parameter.
    """
    if param.startswith('"') and param.endswith('"'):
        return param[1:-1].replace('""', '"')

    return param


def encode_string_param(param: str, force_quotes: bool = False) -> str:
    """Encode a string parameter for sending to the Host Command service.

    Wraps the string in double quotes if necessary, and escapes double quotes.

    Args:
        param: The parameter to encode.
        force_quotes: Whether to force the string to be wrapped in double quotes.

    Returns:
        The encoded parameter.
    """
    if '"' in param or " " in param or force_quotes:
        param = param.replace('"', '""')
        return f'"{param}"'

    return param


def parse_byte_param(byte_string: str) -> bytearray:
    """Convert a "bytes" parameter string to a byte array.

    "Bytes" parameters are sent as a string of signed 32-bit integers, separated by
    commas or spaces, and wrapped in curly or square brackets.

    Args:
        byte_string: The byte array parameter, as a string.

    Returns:
        The byte array.
    """
    # Extract all integer tokens from the string
    tokens = [int(x) for x in re.findall(r"-?\d+", byte_string)]

    # Convert each token to a signed 32-bit integer and create a byte array
    byte_array = bytearray()
    for token in tokens:
        signed_int = struct.pack("i", token)
        byte_array.extend(signed_int)

    return byte_array


def encode_byte_param(byte_array: bytearray) -> str:
    """Convert a byte array to a "bytes" parameter string.

    Args:
        byte_array: The byte array to convert.

    Returns:
        The byte array parameter, as a string.
    """
    # Convert each signed 32-bit integer in the byte array to a string token
    tokens = []
    for byte in range(0, len(byte_array), 4):
        signed_int = struct.unpack("i", byte_array[byte : byte + 4])[0]
        tokens.append(str(signed_int))

    # Join the tokens with commas and wrap in curly braces
    data = "{" + ",".join(tokens) + "}"

    return data



"""Interface for querying and controlling sensors."""

from decimal import Decimal
from typing import Sequence

from .base import Interface


class TemperatureInterface(Interface):
    """Interface for querying and controlling sensors."""

    async def get_value(self, vid: int) -> Decimal:
        """Get the value of a temperature sensor.

        Args:
            vid: The Vantage ID of the temperature sensor.
        """
        # INVOKE <id> Temperature.GetValue
        # -> R:INVOKE <id> <temp> Temperature.GetValue
        response = await self.invoke(vid, "Temperature.GetValue")
        level = Decimal(response.args[1])

        return level

    @classmethod
    def parse_get_value_status(cls, args: Sequence[str]) -> Decimal:
        """Parse a 'Temperature.GetValue' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the sensor.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Temperature.GetValue <temp>
        # STATUS ADD <id>
        # -> S:STATUS <id> Temperature.GetValue <temp>
        return Decimal(args[0]) / 1000



"""Base class for command client interfaces."""

from typing import Union

from aiovantage.command_client import CommandClient, CommandResponse


class Interface:
    """Base class for command client object interfaces."""

    def __init__(self, client: CommandClient) -> None:
        """Initialize an object interface for standalone use.

        Args:
            client: The command client to use.
        """
        self._command_client = client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client."""
        return self._command_client

    async def invoke(
        self, vid: int, method: str, *params: Union[int, float, str]
    ) -> CommandResponse:
        """Invoke a method on an object.

        Args:
            vid: The Vantage ID of the object to invoke the method on.
            method: The name of the method to invoke.
            params: The parameters to pass to the method.
        """
        return await self.command_client.command("INVOKE", vid, method, *params)



"""Controller holding and managing Vantage temperature sensors."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import TemperatureInterface
from aiovantage.models import Temperature

from .base import BaseController, State


class TemperatureSensorsController(BaseController[Temperature], TemperatureInterface):
    """Controller holding and managing Vantage temperature sensors."""

    vantage_types = ("Temperature",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("Temperature.GetValue",)
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a temperature sensor."""
        return {
            "value": await TemperatureInterface.get_value(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a temperature sensor."""
        if status != "Temperature.GetValue":
            return None

        return {
            "value": TemperatureInterface.parse_get_value_status(args),
        }



"""Duck typing for methods."""

from typing import ClassVar, Optional, Protocol, TypeVar

Call = TypeVar("Call")
Return = TypeVar("Return")


class Method(Protocol[Call, Return]):
    """Duck typing for methods."""

    interface: ClassVar[str]
    call: Optional[Call]
    return_value: Optional[Return]



"""Controller holding and managing Vantage loads."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import LoadInterface
from aiovantage.models import Load
from aiovantage.query import QuerySet

from .base import BaseController, State


class LoadsController(BaseController[Load], LoadInterface):
    """Controller holding and managing Vantage loads."""

    vantage_types = ("Load",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("LOAD",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a load."""
        return {
            "level": await LoadInterface.get_level(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a load."""
        if status != "LOAD":
            return None

        return {
            "level": LoadInterface.parse_load_status(args),
        }

    @property
    def is_on(self) -> QuerySet[Load]:
        """Return a queryset of all loads that are turned on."""
        return self.filter(lambda load: load.is_on)

    @property
    def is_off(self) -> QuerySet[Load]:
        """Return a queryset of all loads that are turned off."""
        return self.filter(lambda load: not load.is_on)

    @property
    def relays(self) -> QuerySet[Load]:
        """Return a queryset of all loads that are relays."""
        return self.filter(lambda load: load.is_relay)

    @property
    def motors(self) -> QuerySet[Load]:
        """Return a queryset of all loads that are motors."""
        return self.filter(lambda load: load.is_motor)

    @property
    def lights(self) -> QuerySet[Load]:
        """Return a queryset of all loads that are lights."""
        return self.filter(lambda load: load.is_light)



"""IConfiguration.GetObject method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetObject:
    """IConfiguration.GetObject method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional["GetObject.Params"] = field(default=None)
    return_value: Optional[List[ObjectChoice]] = field(
        default=None,
        metadata={
            "name": "Object",
            "wrapper": "return",
        },
    )

    @dataclass
    class Params:
        """IConfiguration.GetObject method parameters."""

        vids: List[int] = field(
            metadata={
                "name": "VID",
            }
        )



"""Child object mixin."""

from dataclasses import dataclass, field


@dataclass
class ChildObject:
    """Child object mixin."""

    @dataclass
    class Parent:
        """Parent tag."""

        id: int

        position: int = field(
            metadata={
                "name": "Position",
                "type": "Attribute",
            }
        )

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    @property
    def parent_id(self) -> int:
        """Return the parent id."""
        return self.parent.id

    @property
    def parent_position(self) -> int:
        """Return the parent position."""
        return self.parent.position



"""Configuration interface (IConfiguration) method definitions."""

from .close_filter import CloseFilter
from .get_filter_results import GetFilterResults
from .get_object import GetObject
from .open_filter import OpenFilter

__all__ = [
    "CloseFilter",
    "GetFilterResults",
    "OpenFilter",
    "GetObject",
]



"""DryContact object."""

from dataclasses import dataclass, field
from typing import Optional

from .child_object import ChildObject
from .location_object import LocationObject


@dataclass
class DryContact(ChildObject, LocationObject):
    """DryContact object."""

    triggered: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Blind object."""

from dataclasses import dataclass

from .blind_base import BlindBase
from .child_object import ChildObject
from .location_object import LocationObject


@dataclass
class Blind(BlindBase, ChildObject, LocationObject):
    """Blind object."""



"""Button object."""

from dataclasses import dataclass, field
from typing import Optional

from .child_object import ChildObject
from .system_object import SystemObject


@dataclass
class Button(ChildObject, SystemObject):
    """Button object."""

    text1: str = field(
        metadata={
            "name": "Text1",
        }
    )

    text2: str = field(
        metadata={
            "name": "Text2",
        }
    )

    up_id: int = field(
        metadata={
            "name": "Up",
        }
    )

    down_id: int = field(
        metadata={
            "name": "Down",
        }
    )

    hold_id: int = field(
        metadata={
            "name": "Hold",
        }
    )

    pressed: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""ILogin.Login method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional


@dataclass
class Login:
    """ILogin.Login method definition."""

    interface: ClassVar[str] = "ILogin"

    call: Optional["Login.Params"] = field(default=None)

    return_value: Optional[bool] = field(
        default=None,
        metadata={
            "name": "return",
        },
    )

    @dataclass
    class Params:
        """ILogin.Login method parameters."""

        user: str = field(
            metadata={
                "name": "User",
            }
        )

        password: str = field(
            metadata={
                "name": "Password",
            }
        )



"""Interface for keypad speakers."""

from .base import Interface


class SounderInterface(Interface):
    """Interface for keypad speakers."""

    async def turn_on(self, vid: int) -> None:
        """Turn on the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
        """
        # INVOKE <id> Sounder.TurnOn
        # -> R:INVOKE <id> <rcode> Sounder.SetStatus
        await self.invoke(vid, "Sounder.SetStatus", "On")

    async def turn_off(self, vid: int) -> None:
        """Turn off the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
        """
        # INVOKE <id> Sounder.TurnOff
        # -> R:INVOKE <id> <rcode> Sounder.SetStatus
        await self.invoke(vid, "Sounder.SetStatus", "Off")

    async def get_frequency(self, vid: int) -> float:
        """Get the frequency of the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.

        Returns:
            The frequency of the keypad speaker.
        """
        # INVOKE <id> Sounder.GetFrequency
        # -> R:INVOKE <id> <frequency> Sounder.GetFrequency
        response = await self.invoke(vid, "Sounder.GetFrequency")
        frequency = float(response.args[0])

        return frequency

    async def set_frequency(self, vid: int, frequency: float) -> None:
        """Set the frequency of the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
            frequency: The frequency to set.
        """
        # INVOKE <id> Sounder.SetFrequency <frequency>
        # -> R:INVOKE <id> <rcode> Sounder.SetFrequency
        await self.invoke(vid, "Sounder.SetFrequency", frequency)

    async def play_fx(
        self, vid: int, effect: int, duration: float = 0, volume: float = 0
    ) -> None:
        """Play a sound effect on the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
            effect: The effect to play.
            duration: The duration to play the FX for, in seconds, 0 for default.
            volume: The volume to play the FX at, as a percentage, 0 for default.
        """
        # INVOKE <id> Sounder.PlayFX <fx> <duration> <volume>
        # -> R:INVOKE <id> <rcode> Sounder.PlayFX
        await self.invoke(vid, "Sounder.PlayFX", effect, duration, volume)



"""Interface for querying and controlling light sensors."""

from decimal import Decimal
from typing import Sequence

from .base import Interface


class LightSensorInterface(Interface):
    """Interface for querying and controlling light sensors."""

    async def get_level(self, vid: int) -> Decimal:
        """Get the level of a light sensor, in foot-candles.

        Args:
            vid: The Vantage ID of the light sensor.
        """
        # INVOKE <id> LightSensor.GetLevel
        # -> R:INVOKE <id> <level> LightSensor.GetLevel
        response = await self.invoke(vid, "LightSensor.GetLevel")
        level = Decimal(response.args[1])

        return level

    @classmethod
    def parse_get_level_status(cls, args: Sequence[str]) -> Decimal:
        """Parse a 'LightSensor.GetLevel' event.

        Args:
            args: The arguments of the event.

        Returns:
            The level of the light sensor, in foot-candles.
        """
        # ELLOG STATUS ON
        # -> EL: <id> LightSensor.GetLevel <level>
        # STATUS ADD <id>
        # -> S:STATUS <id> LightSensor.GetLevel <level>
        return Decimal(args[0]) / 1000



"""Somfy URTSI 2 blind group."""

from dataclasses import dataclass

from .blind_group_base import BlindGroupBase
from .child_device import ChildDevice


@dataclass
class SomfyURTSI2GroupChild(BlindGroupBase, ChildDevice):
    """Somfy URTSI 2 blind group."""

    class Meta:
        name = "Somfy.URTSI_2_Group_CHILD"



"""Interface for querying and controlling blinds."""

from collections import namedtuple
from decimal import Decimal
from typing import Sequence

from .base import Interface


class BlindInterface(Interface):
    """Interface for querying and controlling blinds."""

    BlindState = namedtuple(
        "BlindState",
        ["is_moving", "start_pos", "end_pos", "transition_time", "start_time"],
    )

    async def open(self, vid: int) -> None:
        """Open a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Open
        # -> R:INVOKE <id> <rcode> Blind.Open
        await self.invoke(vid, "Blind.Open")

    async def close(self, vid: int) -> None:
        """Close a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Close
        # -> R:INVOKE <id> <rcode> Blind.Close
        await self.invoke(vid, "Blind.Close")

    async def stop(self, vid: int) -> None:
        """Stop a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Stop
        # -> R:INVOKE <id> <rcode> Blind.Stop
        await self.invoke(vid, "Blind.Stop")

    async def get_position(self, vid: int) -> Decimal:
        """Get the position of a blind.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The position of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetPosition
        # -> R:INVOKE <id> <position (0-100.000)> Blind.GetPosition
        response = await self.invoke(vid, "Blind.GetPosition")
        position = Decimal(response.args[1])

        return position

    async def set_position(self, vid: int, position: float) -> None:
        """Set the position of a blind.

        Args:
            vid: The Vantage ID of the blind.
            position: The position to set the blind to, as a percentage.
        """
        # INVOKE <id> Blind.SetPosition <position>
        # -> R:INVOKE <id> <rcode> Blind.SetPosition <position>
        await self.invoke(vid, "Blind.SetPosition", position)

    @classmethod
    def parse_blind_status(cls, args: Sequence[str]) -> Decimal:
        """Parse a simple 'S:BLIND' event.

        Args:
            args: The arguments of the event.

        Returns:
            The position of the blind, as a percentage.
        """
        # STATUS BLIND
        # -> S:BLIND <id> <position (0-100.000)>
        return Decimal(args[0])

    @classmethod
    def parse_get_position_status(cls, args: Sequence[str]) -> Decimal:
        """Parse a 'Blind.GetPosition' event."""
        # ELLOG STATUS ON
        # -> EL: <id> Blind.GetPosition <position (0-100000)>
        # ADDSTATUS <id>
        # -> S:STATUS <id> Blind.GetPosition <position (0-100000)>
        position = Decimal(args[0]) / 1000

        return position

    @classmethod
    def parse_get_state_status(cls, args: Sequence[str]) -> BlindState:
        """Parse a 'Blind.GetBlindState' event."""
        # ADDSTATUS <id>
        # -> S:STATUS <id> Blind.GetBlindState <moving> <start> <end> <transitionTime> <startTime>
        # ELLOG STATUSEX ON
        # -> EL: <id> Blind.GetBlindState <moving> <start> <end> <transitionTime> <startTime>
        return cls.BlindState(
            # Is the blind currently moving?
            is_moving=bool(int(args[0])),
            # Position the blind is moving from (as a percentage)
            start_pos=Decimal(args[1]) / 1000,
            # Position the blind is moving to (as a percentage)
            end_pos=Decimal(args[2]) / 1000,
            # Time the blind will take to move (in seconds)
            transition_time=Decimal(args[3]) / 1000,
            # Time the blind started moving (in milliseconds since start of UTC day)
            start_time=int(args[4]),
        )



"""Interface for querying and controlling system objects."""

from typing import Sequence

from .base import Interface


class ObjectInterface(Interface):
    """Interface for querying and controlling system objects."""

    async def get_mtime(self, vid: int) -> int:
        """Get the modification time of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The modification time of the object, as a unix timestamp.
        """
        # INVOKE <id> Object.GetMTime
        # -> R:INVOKE <id> <mtime> Object.GetMTime
        response = await self.invoke(vid, "Object.GetMTime")
        mtime = int(response.args[1])

        return mtime

    @classmethod
    def parse_get_mtime_status(cls, args: Sequence[str]) -> int:
        """Parse an 'Object.GetMTime' event.

        Args:
            args: The arguments of the event.

        Returns:
            The modification time of the object, as a unix timestamp.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Object.GetMTime <mtime>
        # STATUS ADD <id>
        # -> S:STATUS <id> Object.GetMTime <mtime>
        return int(args[0])



"""Controller holding and managing Vantage RGB loads."""

from typing import Any, Dict, List, Optional, Sequence, Tuple

from typing_extensions import override

from aiovantage.command_client.interfaces import (
    ColorTemperatureInterface,
    LoadInterface,
    RGBLoadInterface,
)
from aiovantage.models import RGBLoadBase
from aiovantage.query import QuerySet

from .base import BaseController, State


class RGBLoadsController(
    BaseController[RGBLoadBase],
    LoadInterface,
    RGBLoadInterface,
    ColorTemperatureInterface,
):
    """Controller holding and managing Vantage RGB loads."""

    vantage_types = ("Vantage.DGColorLoad", "Vantage.DDGColorLoad")
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = (
        "RGBLoad.GetHSL",
        "RGBLoad.GetRGB",
        "RGBLoad.GetRGBW",
        "ColorTemperature.Get",
        "Load.GetLevel",
    )
    """Which status methods this controller handles from the Enhanced Log."""

    def __post_init__(self) -> None:
        """Initialize the map for building colors."""
        self._temp_color_map: Dict[int, List[int]] = {}

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of an RGB load."""
        state: Dict[str, Any] = {
            "level": await LoadInterface.get_level(self, vid),
        }

        rgb_load: RGBLoadBase = self[vid]
        if rgb_load.is_rgb:
            state["hsl"] = await RGBLoadInterface.get_hsl(self, vid)
            state["rgb"] = await RGBLoadInterface.get_rgb(self, vid)
            state["rgbw"] = await RGBLoadInterface.get_rgbw(self, vid)

        if rgb_load.is_cct:
            state["color_temp"] = await ColorTemperatureInterface.get_color_temp(
                self, vid
            )

        return state

    @override
    def parse_object_update(self, vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for an RGB load."""
        rgb_load: RGBLoadBase = self[vid]
        if status == "Load.GetLevel":
            return {
                "level": LoadInterface.parse_get_level_status(args),
            }

        if status == "RGBLoad.GetHSL" and rgb_load.is_rgb:
            channel, value = RGBLoadInterface.parse_color_channel_status(args)
            if hsl := self._build_color_from_channels(vid, channel, value, 3):
                return {"hsl": hsl}
            return None

        if status == "RGBLoad.GetRGB" and rgb_load.is_rgb:
            channel, value = RGBLoadInterface.parse_color_channel_status(args)
            if rgb := self._build_color_from_channels(vid, channel, value, 3):
                return {"rgb": rgb}
            return None

        if status == "RGBLoad.GetRGBW" and rgb_load.is_rgb:
            channel, value = RGBLoadInterface.parse_color_channel_status(args)
            if rgbw := self._build_color_from_channels(vid, channel, value, 4):
                return {"rgbw": rgbw}
            return None

        if status == "ColorTemperature.Get" and rgb_load.is_cct:
            return {
                "color_temp": ColorTemperatureInterface.parse_get_status(args),
            }

        return None

    @property
    def is_on(self) -> QuerySet[RGBLoadBase]:
        """Return a queryset of all RGB loads that are turned on."""
        return self.filter(lambda load: load.is_on)

    @property
    def is_off(self) -> QuerySet[RGBLoadBase]:
        """Return a queryset of all RGB loads that are turned off."""
        return self.filter(lambda load: not load.is_on)

    def _build_color_from_channels(
        self, vid: int, channel: int, value: int, num_channels: int
    ) -> Optional[Tuple[int, ...]]:
        # Build a color from a series of channel value events. We need to store
        # partially constructed colors in memory, since updates come separately for
        # each channel.

        # Ignore updates for channels we don't care about
        if channel < 0 or channel >= num_channels:
            return None

        # Store the channel value in the temp color map
        self._temp_color_map.setdefault(vid, num_channels * [0])
        self._temp_color_map[vid][channel] = value

        # If we have all the channels, build and return the color
        if channel == num_channels - 1:
            color = tuple(self._temp_color_map[vid])
            del self._temp_color_map[vid]
            return color

        return None



"""Run a task by id."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage import Vantage

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("id", help="task id to run")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller and print out the name and value of each GMem
    async with Vantage(args.host, args.username, args.password) as vantage:
        try:
            task_id = int(args.id)
        except ValueError:
            print("Invalid task id")
            return

        task = await vantage.tasks.aget(task_id)
        if task is None:
            print("Task not found")
            return

        print(f"{task.name} (id = {task.id})")
        await vantage.tasks.start(task.id)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""IConfiguration.GetFilterResults method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetFilterResults:
    """IConfiguration.GetFilterResults method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional["GetFilterResults.Params"] = field(default=None)
    return_value: Optional[List[ObjectChoice]] = field(
        default=None,
        metadata={
            "name": "Object",
            "wrapper": "return",
        },
    )

    @dataclass
    class Params:
        """IConfiguration.GetFilterResults method parameters."""

        h_filter: int = field(
            metadata={
                "name": "hFilter",
            }
        )

        count: int = field(
            default=50,
            metadata={
                "name": "Count",
            },
        )

        whole_object: bool = field(
            default=True,
            metadata={
                "name": "WholeObject",
            },
        )



"""Controller holding and managing Vantage stations."""

from aiovantage.models import StationObject

from .base import BaseController


class StationsController(BaseController[StationObject]):
    """Controller holding and managing Vantage stations."""

    vantage_types = (
        "Dimmer",
        "DualRelayStation",
        "EqCtrl",
        "EqUX",
        "Keypad",
        "ScenePointRelay",
        "Vantage.DmxDaliGateway",
    )
    """The Vantage object types that this controller will fetch."""



"""AnemoSensor (wind sensor) object."""

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Optional

from .child_object import ChildObject
from .sensor import Sensor


@dataclass
class AnemoSensor(ChildObject, Sensor):
    """AnemoSensor (wind sensor) object."""

    speed: Optional[Decimal] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Load object."""

from dataclasses import dataclass, field
from typing import Optional

from .child_object import ChildObject
from .location_object import LocationObject


@dataclass
class Load(ChildObject, LocationObject):
    """Load object."""

    load_type: str = field(
        metadata={
            "name": "LoadType",
        }
    )

    power_profile_id: int = field(
        metadata={
            "name": "PowerProfile",
        }
    )

    level: Optional[float] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def is_relay(self) -> bool:
        """Return True if the load is a relay."""
        return self.load_type in (
            "High Voltage Relay",
            "Low Voltage Relay",
        )

    @property
    def is_motor(self) -> bool:
        """Return True if the load is a motor."""
        return self.load_type == "Motor"

    @property
    def is_light(self) -> bool:
        """Return True if the load is a light."""
        return not (self.is_relay or self.is_motor)

    @property
    def is_dimmable(self) -> bool:
        """Return True if the load is dimmable."""
        return not self.load_type.endswith("non-Dim")

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



"""Interface for querying and controlling buttons."""

from typing import Sequence

from .base import Interface


class ButtonInterface(Interface):
    """Interface for querying and controlling buttons."""

    async def get_state(self, vid: int) -> bool:
        """Get the state of a button.

        Args:
            vid: The Vantage ID of the button.

        Returns:
            The pressed state of the button, True if pressed, False if not.
        """
        # INVOKE <id> Button.GetState
        # -> R:INVOKE <id> <state (Up/Down)> Button.GetState
        response = await self.invoke(vid, "Button.GetState")
        state = response.args[1]
        if state == "Up":
            return False
        if state == "Down":
            return True

        raise ValueError(f"Invalid button state name: {state}")

    async def set_state(self, vid: int, pressed: bool) -> None:
        """Set the state of a button.

        Args:
            vid: The Vantage ID of the button.
            pressed: The state to set the button to, either a State.UP or State.DOWN.
        """
        # INVOKE <id> Button.SetState <state (0/1/Up/Down)>
        # -> R:INVOKE <id> Button.SetState <state (Up/Down)>
        await self.invoke(vid, "Button.SetState", pressed)

    async def press(self, vid: int) -> None:
        """Press a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.set_state(vid, True)

    async def release(self, vid: int) -> None:
        """Release a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.set_state(vid, False)

    async def press_and_release(self, vid: int) -> None:
        """Press and release a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.press(vid)
        await self.release(vid)

    @classmethod
    def parse_btn_status(cls, args: Sequence[str]) -> bool:
        """Parse a simple 'S:BTN' event.

        Args:
            args: The arguments of the event.

        Returns:
            The state of the button, either a State.UP or State.DOWN.
        """
        # STATUS BTN
        # -> S:BTN <id> <state (PRESS/RELEASE)>
        state = args[0]
        if state == "RELEASE":
            return False
        if state == "PRESS":
            return True

        raise ValueError(f"Invalid button state name: {state}")

    @classmethod
    def parse_get_state_status(cls, args: Sequence[str]) -> bool:
        """Parse a 'Button.GetState' event.

        Args:
            args: The arguments of the event.

        Returns:
            The state of the button, either a State.UP or State.DOWN.
        """
        # ELLOG STATUS ON
        # -> EL: <id> Button.GetState <state (0/1)>
        # STATUS ADD <id>
        # -> S:STATUS <id> Button.GetState <state (0/1)>
        return bool(int(args[0]))



"""Interface for controller introspection."""

from enum import IntEnum

from .base import Interface


class IntrospectionInterface(Interface):
    """Interface for controller introspection."""

    class Firmware(IntEnum):
        """Firmware images."""

        KERNEL = 0
        ROOT_FS = 1
        APPLICATION = 2

    async def get_firmware_version(self, vid: int, image: Firmware) -> str:
        """Get the firmware version.

        Args:
            vid: The Vantage ID of the controller.
            image: The firmware image to get the version of.
        """
        # INVOKE <id> Introspection.GetFirmwareVersion <image>
        # -> R:INVOKE <id> <rcode> Introspection.GetFirmwareVersion <image> <version>
        response = await self.invoke(vid, "Introspection.GetFirmwareVersion", image)

        return response.args[4].rstrip()



"""IConfiguration.CloseFilter method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional


@dataclass
class CloseFilter:
    """IConfiguration.CloseFilter method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional[int] = field(default=None)
    return_value: Optional[bool] = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""ModuleGen2 object, eg. SDM12-EM, UDM08-EM."""

from dataclasses import dataclass

from .child_object import ChildObject
from .system_object import SystemObject


@dataclass
class ModuleGen2(ChildObject, SystemObject):
    """ModuleGen2 object, eg. SDM12-EM, UDM08-EM."""



"""BlindGroup object."""

from dataclasses import dataclass, field
from typing import List

from .blind_group_base import BlindGroupBase
from .location_object import LocationObject


@dataclass
class BlindGroup(BlindGroupBase, LocationObject):
    """BlindGroup object."""

    blind_ids: List[int] = field(
        default_factory=list,
        metadata={
            "name": "Blind",
            "wrapper": "BlindTable",
        },
    )



"""Controller holding and managing Vantage anemo (wind) sensors."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.interfaces import AnemoSensorInterface
from aiovantage.models import AnemoSensor

from .base import BaseController, State


class AnemoSensorsController(BaseController[AnemoSensor], AnemoSensorInterface):
    """Controller holding and managing Vantage anemo (wind) sensors."""

    vantage_types = ("AnemoSensor",)
    """The Vantage object types that this controller will fetch."""

    enhanced_log_status_methods = ("AnemoSensor.GetSpeed",)
    """Which status methods this controller handles from the Enhanced Log."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of an anemo sensor."""
        return {
            "speed": await AnemoSensorInterface.get_speed(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for an anemo sensor."""
        if status != "AnemoSensor.GetSpeed":
            return None

        return {
            "speed": AnemoSensorInterface.parse_get_speed_status(args),
        }



"""Fetch all buttons from the Vantage controller, and print out any state changes."""

import argparse
import asyncio
import contextlib
import logging
from typing import Any, Dict

from aiovantage import Vantage, VantageEvent
from aiovantage.models import Button

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def callback(event: VantageEvent, obj: Button, data: Dict[str, Any]) -> None:
    """Print out any state changes."""
    if event == VantageEvent.OBJECT_ADDED:
        print(f"[Button added] '{obj.name}' ({obj.id})")

    elif event == VantageEvent.OBJECT_UPDATED:
        print(f"[Button updated] '{obj.name}' ({obj.id})")
        for attr in data.get("attrs_changed", []):
            print(f"    {attr} = {getattr(obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Subscribe to updates for all buttons
        vantage.buttons.subscribe(callback)

        # Fetch all known buttons from the controller
        await vantage.buttons.initialize()

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""GMem (variable) object."""

from dataclasses import dataclass, field
from typing import Union

from .system_object import SystemObject


@dataclass
class Tag:
    """GMem tag."""

    type: str

    object: bool = field(
        default=False,
        metadata={
            "name": "object",
            "type": "Attribute",
        },
    )


@dataclass
class Data:
    """GMem data."""

    fixed: bool = field(
        default=False,
        metadata={
            "name": "Fixed",
            "type": "Attribute",
        },
    )


@dataclass
class GMem(SystemObject):
    """GMem (variable) object."""

    data: Data = field(
        metadata={
            "name": "data",
        }
    )

    persistent: bool = field(
        metadata={
            "name": "Persistent",
        }
    )

    tag: Tag = field(
        metadata={
            "name": "Tag",
        }
    )

    value: Union[int, str, bool, None] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def is_bool(self) -> bool:
        """Return True if GMem is boolean type."""
        return self.tag.type == "bool"

    @property
    def is_str(self) -> bool:
        """Return True if GMem is string type."""
        return self.tag.type == "Text"

    @property
    def is_int(self) -> bool:
        """Return True if GMem is integer type."""
        return self.tag.type in (
            "Delay",
            "DeviceUnits",
            "Level",
            "Load",
            "Number",
            "Seconds",
            "Task",
            "DegC",
        )

    @property
    def is_object_id(self) -> bool:
        """Return True if GMem is a pointer to an object."""
        return self.tag.object

    @property
    def is_fixed(self) -> bool:
        """Return True if GMem is a fixed point number."""
        return self.data.fixed



"""Send commands to the Vantage Host Command service."""

import asyncio
import logging
from dataclasses import dataclass
from decimal import Decimal
from ssl import SSLContext
from types import TracebackType
from typing import List, Optional, Type, Union

from typing_extensions import Self

from aiovantage.connection import BaseConnection
from aiovantage.errors import CommandError, LoginFailedError, LoginRequiredError

from .utils import encode_params, tokenize_response


class CommandConnection(BaseConnection):
    """Connection to a Vantage Host Command service."""

    default_port = 3001
    default_ssl_port = 3010


@dataclass
class CommandResponse:
    """Wrapper for command responses."""

    command: str
    args: List[str]
    data: List[str]

    def __init__(self, data: List[str]) -> None:
        """Initialize a CommandResponse."""
        self.data, return_line = data[:-1], data[-1]
        command, *self.args = tokenize_response(return_line)
        self.command = command[2:]


@dataclass
class InvokeResponse:
    """Wrapper for invoke command responses."""

    vid: int
    result: str
    method: str
    args: List[str]

    def __init__(self, data: List[str]) -> None:
        """Initialize an InvokeResponse."""
        _, vid_str, self.result, self.method, *self.args = tokenize_response(data[0])
        self.vid = int(vid_str)


class CommandClient:
    """Client to send commands to the Vantage Host Command service."""

    def __init__(
        self,
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        *,
        ssl: Union[SSLContext, bool] = True,
        port: Optional[int] = None,
        conn_timeout: float = 30,
        read_timeout: float = 60,
    ) -> None:
        """Initialize the client."""
        self._connection = CommandConnection(host, port, ssl, conn_timeout)
        self._username = username
        self._password = password
        self._read_timeout = read_timeout
        self._connection_lock = asyncio.Lock()
        self._command_lock = asyncio.Lock()
        self._logger = logging.getLogger(__name__)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit context manager."""
        self.close()
        if exc_val:
            raise exc_val

    def close(self) -> None:
        """Close the connection to the Host Command service."""
        self._connection.close()

    async def command(
        self,
        command: str,
        *params: Union[str, int, float, Decimal],
        force_quotes: bool = False,
        connection: Optional[CommandConnection] = None,
    ) -> CommandResponse:
        """Send a command to the Host Command service and wait for a response.

        Args:
            command: The command to send, should be a single word string.
            params: The parameters to send with the command.
            force_quotes: Whether to force string params to be wrapped in double quotes.
            connection: The connection to use, if not the default.

        Returns:
            A CommandResponse instance.
        """
        if params:
            request = f"{command} {encode_params(*params, force_quotes=force_quotes)}"
        else:
            request = command

        # Send the request and parse the response
        raw_response = await self.raw_request(request, connection=connection)
        return CommandResponse(raw_response)

    async def invoke(
        self,
        vid: int,
        method: str,
        *params: Union[str, int, float, Decimal],
        force_quotes: bool = False,
        connection: Optional[CommandConnection] = None,
    ) -> InvokeResponse:
        """Invoke a method on an object, and wait for a response.

        Args:
            vid: The VID of the object to invoke the command on.
            method: The method to invoke.
            params: The parameters to send with the method.
            force_quotes: Whether to force string params to be wrapped in double quotes.
            connection: The connection to use, if not the default.

        Returns:
            A InvokeResponse instance.
        """
        if params:
            request = f"INVOKE {vid} {method} {encode_params(*params, force_quotes=force_quotes)}"
        else:
            request = f"INVOKE {vid} {method}"

        # Send the request and parse the response
        raw_response = await self.raw_request(request, connection=connection)
        return InvokeResponse(raw_response)

    async def raw_request(
        self, request: str, connection: Optional[CommandConnection] = None
    ) -> List[str]:
        """Send a raw command to the Host Command service and return all response lines.

        Handles authentication if required, and raises an exception if the response line
        contains R:ERROR.

        Args:
            request: The request to send.
            connection: The connection to use, if not the default.

        Returns:
            The response lines received from the server.
        """
        conn = connection or await self.get_connection()

        # Send the command
        async with self._command_lock:
            self._logger.debug("Sending command: %s", request)
            await conn.write(f"{request}\n")

            # Read all lines of the response
            response_lines = []
            while True:
                response_line = await conn.readuntil(b"\r\n", self._read_timeout)
                response_line = response_line.rstrip()

                # Handle error codes
                if response_line.startswith("R:ERROR"):
                    raise self._parse_command_error(response_line)

                # Ignore potentially interleaved "event" messages
                if any(response_line.startswith(x) for x in ("S:", "L:", "EL:")):
                    self._logger.debug("Ignoring event message: %s", response_line)
                    continue

                # Return the response once we see the response line
                response_lines.append(response_line)
                if response_line.startswith("R:"):
                    break

        self._logger.debug("Received response: %s", "\n".join(response_lines))

        return response_lines

    async def get_connection(self) -> CommandConnection:
        """Get a connection to the Host Command service."""
        async with self._connection_lock:
            if self._connection.closed:
                # Open a new connection
                await self._connection.open()

                # Authenticate the new connection if we have credentials
                if self._username and self._password:
                    await self.command(
                        "LOGIN",
                        self._username,
                        self._password,
                        connection=self._connection,
                    )

            return self._connection

    def _parse_command_error(self, message: str) -> CommandError:
        # Parse a command error from a message.
        tag, error_message = message.split(" ", 1)
        _, _, error_code_str = tag.split(":")
        error_code = int(error_code_str)

        exc: CommandError
        if error_code == 21:
            exc = LoginRequiredError(error_message)
        elif error_code == 23:
            exc = LoginFailedError(error_message)
        else:
            exc = CommandError(f"{error_message} (Error code {error_code})")

        return exc
